#!/usr/bin/env python3
import zlib
import argparse
import sys
import os
from functools import reduce
from PIL import Image

def split_to_lines(data, line_size, assert_aligned=True):
    if assert_aligned:
        assert len(data) % line_size == 0
    return [data[i:i+line_size] for i in range(0, len(data), line_size)]

def extract_raw_png_data(fdata, debug=False):
        data = b''
        offset = fdata.find(b'IDAT')
        if offset == -1:
            print('[-] IDAT not found')
            return 
        offset -= 4
        begin_offset = offset

        while True:
            idat_len = int.from_bytes(fdata[offset:offset+4], byteorder='big')
            idat_marker = fdata[offset+4:offset+8]
            if idat_marker != b'IDAT':
                if debug:
                    print('[+] Stopping at {}'.format(idat_marker))
                break
            if debug:
                print('[+] Found IDAT chunk at {}, size {}'.format(hex(offset), hex(idat_len)))
            data += fdata[offset+8:offset+8+idat_len]
            offset += idat_len + 12
        if debug:
            print('[+] Raw data size: {}'.format(hex(len(data)))) 
        return data, begin_offset, offset

def extract_png_data(fdata, debug=False):
    raw_data, begin_offset, end_offset = extract_raw_png_data(fdata, debug=debug)
    data = zlib.decompress(raw_data)
    if debug:
        print('[+] Decompressed size: {}'.format(hex(len(data))))
    return data, begin_offset, end_offset  

def save_png(src_png_path, dest_png_path, data, begin_offset, end_offset, debug=False):
    with open(src_png_path, 'rb') as f:
        src_data = f.read()
    data = zlib.compress(data)
    if debug:
        print('[+] Compressed size: {}'.format(len(data)))
    idat_len = len(data).to_bytes(4, byteorder='big')
    idat = b'IDAT' + data
    idat_crc = zlib.crc32(idat).to_bytes(4, byteorder='big')
    idat = idat_len + idat + idat_crc
    w_data = src_data[:begin_offset] + idat + src_data[end_offset:]
    with open(dest_png_path, 'wb') as f:
        f.write(w_data)

def filter_encode(line, bpp):
    raw = bytearray(line)
    sub = raw[::]
    for x in range(len(line)):
        m_idx = x - bpp
        if m_idx < 0:
            sub_val = 0
        else:
            sub_val = raw[m_idx]
        sub[x] = (raw[x] - sub_val) % 256

    return bytes(sub)

class PNG(object):
    def __init__(self, path):
        with open(path, 'rb') as f:
            data = f.read()
        self.path = path
        self.data = data
        offset = data.find(b'IHDR')
        self.w = int.from_bytes(data[offset+4:offset+8], byteorder='big')
        self.h = int.from_bytes(data[offset+8:offset+12], byteorder='big')
        bpp = data[offset+13]
        if bpp not in [0, 2, 4, 6]:
            print('[-] Color type not supported, exiting')
            os.exit(1)
        else:
            self.bpp = {
                0: 1,
                2: 3,
                4: 2,
                6: 4,
            }[bpp]

        self.bitmap, self.begin_offset, self.end_offset = extract_png_data(data)

    def encode(self, data, debug=False):
        res = b''
        bin_data = bin(int.from_bytes(data, byteorder='big'))[2:]
        bin_data = bin_data.rjust(((len(bin_data)+7)//8)*8, '0')
        img = Image.open(self.path).tobytes()
        img_lines = split_to_lines(img, self.w*self.bpp)
        for idx, line in enumerate(img_lines):
            if idx < len(bin_data):
                if bin_data[idx] == '0':
                    res += b'\x00'
                elif bin_data[idx] == '1':
                    res += b'\x01'
                    line = filter_encode(line, self.bpp)
            else:
                res += b'\x00'
            res += line

        if debug:
            print('[+] Encoded data size: {}'.format(hex(len(res))))

        self.raw_data = res
        res = zlib.compress(res)
        if debug:
            print('[+] Compressed size: {}'.format(len(res)))
        idat_len = len(res).to_bytes(4, byteorder='big')
        idat = b'IDAT' + res
        idat_crc = zlib.crc32(idat).to_bytes(4, byteorder='big')
        idat = idat_len + idat + idat_crc
        self.data = self.data[:self.begin_offset] + idat + self.data[self.end_offset:]

    def decode(self):
        lines = split_to_lines(self.bitmap, self.w*self.bpp + 1)
        data = ''.join([str(line[0] & 1) for line in lines])
        data = int(data, 2)
        return data.to_bytes((data.bit_length() + 7) // 8, byteorder='big')

    def save(self, path):
        with open(path, 'wb') as f:
            f.write(self.data) 


def main():
    parser = argparse.ArgumentParser()
    opt_group = parser.add_mutually_exclusive_group()
    opt_group.add_argument("-d", "--decode", action="store_true",
                    help="decode data from image")
    opt_group.add_argument("-e", "--encode", action="store_true",
                    help="encode data into image")
    parser.add_argument('-dt', '--data', type=str,
                    help='path to data you want to encode')
    parser.add_argument('input', type=str,
                    help='path to the input file')
    parser.add_argument('-out', '--output', type=str,
                    help='path to the output file', default='stdout')
    args = parser.parse_args()
    png = PNG(args.input)
    if args.decode:
        dec_data = png.decode()
        if args.output == 'stdout':
            sys.stdout.write(str(dec_data))
        else:
            with open(args.output, 'wb') as f:
                f.write(dec_data)
    elif args.encode:
        with open(args.data, 'rb') as f:
            data_to_enc = f.read()
        png.encode(data_to_enc)
        png.save(args.output)

if __name__ == '__main__':
    main()